# cry_h1.py
# The Cryptanalysis of the Single Cipher Machine
# Author: JF Bouchaudy, 7/2023
# This software follows specifictions of Don Dawson's software
# written in Basic in 9/1985.

import sys
import string

# ====> Global variables <===================
ALPHA = string.ascii_uppercase
MAX  = 5000
CRYPTO = [0]*MAX
N = 26
LG = -1
tb = [0]*N
tb2 = [-1]*N
REVERSE = False
DIAGONALS = []
MATRIX = []
RED_TABLE = []
for i in range(N):
    DIAGONALS.append(list(tb))
    MATRIX.append(list(tb))
    RED_TABLE.append(list(tb))
FLAG_MATRIX = False
DIAGONAL_TOTAL = list(tb)
FLAG_BASE = False
COMBINED = {
    'letter': list(tb2),
    'stats': list(tb),
    'nblet': 0,
    'histo': []
}
DECAL_HYPO = list(tb2)

#--------------------------------------------
def menu():
    print("1 - Read the cryptogram from a file")
    print("2 - Print the cryptogram")
    print("3 - Choose the Reverse mode (Standard mode by default)")
    print("4 - Create matrix and diagonals")
    print("5 - Print matrix")
    print("6 - Print diagonals")
    print("7 - Print phi values of diagonals")
    print("8 - Add an alphabet (the first will be the base)")
    print("9 - Calculate PHI for remaining alphabets")
    print("A - Print combined alphabet (or any alphabet)")
    print("B - Constructing the Reduction Tableau")
    print("C - Combine all alphabets")
    print("D - Match two diagonals")
    print("E - Reset the combined alphabet")
    print("F - Reduction to a Monoalphabet Ciphertext")
#--------------------------------------------
def num2let(un_tb):
    alphabet = ""
    for i in range(N):
        if un_tb[i] == -1:
            alphabet += "."
        else:
            alphabet += ALPHA[un_tb[i]]
    return alphabet
#--------------------------------------------
def num_of_letters(row):
    total = 0
    for i in range(N):
        total += DIAGONALS[row][i]
    return total
#--------------------------------------------
def calculate_phi(tb):
    """ """ 
    total = 0
    for i in range(N):
        if tb[i] > 1:
            total += tb[i] * (tb[i] - 1)
    return total 

#--------------------------------------------
def valide_num_26(ch):
    if ch.isnumeric():
        val = int(ch)
        if val >= 0 and val < N:
            return True
    return False
#--------------------------------------------
def lecture():
    """ """
    global LG
    if LG != -1:
        print("ERROR: You need to re-run the software to read", end="")
        print(" another cryptogram (sorry :-(")
        return
    nom = input("File name ? ")
    try:
        f = open(nom)    
        ch = f.read()
        if len(ch) >= MAX:
            print("ERROR: Cryptogram too long")
            sys.exit(0)
        f.close()
    except:
        print("ERROR: Impossible to open cryptogram file")
        return
    else:
        i = 0
        for c in ch:
            k = ALPHA.find(c.upper())
            if k >= 0: 
                CRYPTO[i] = k
                i += 1
        LG = i
#--------------------------------------------
def print_cry():
    """ """
    if LG == -1:
        print("ERROR: You need to read a cryptogram first")
        return
    for i in range(LG):
        if (i%N) == 0: print()
        print("%c " % (chr(CRYPTO[i]+ord('A'))),end="")
    print()
    print("Cryptogram length: ", LG)
#--------------------------------------------
def reverse_mode():
    """ """
    global REVERSE

    print("Fix the Reverse mode ...")
    print("Choose the Reverse mode (return to keep current value)")
    ch = input("      (R: Reverse, S: Standard): ? ")
    if ch == "": 
        pass
    elif ch == "R":
        REVERSE = True
    elif ch == "S":
        REVERSE = False
    else:
        print("ERROR (keep current value)")
    mode = "Reverse" if REVERSE else "Standard"
    print("Current value: ", mode)
#--------------------------------------------
def create_diagonals():
    """ """
    global FLAG_MATRIX

    # 1. test if there is a cryptogam
    if LG == -1:
        print("ERROR: You need to read a cryptogram first")
        return
    print("Create Matrix and Diagonals...")

    # 2. reset MATRIX and DIAGONALS arrays and number of letters
    for i in range(N):
        DIAGONAL_TOTAL[i] = 0
        for j in range(N):
            MATRIX[i][j] = 0
            DIAGONALS[i][j] = 0

    # 3. Calculate Matrix array
    for k in range(LG):
        col = k % N
        row = CRYPTO[k]
        MATRIX[col][row] += 1

    # 4. Calculate Diagonals arrays
    for row in range(N):
        for i in range(N):
            if REVERSE:
                j = row + 1
                if j >= N:
                    j -= N
            else:
                j = row - i
                if j < 0:
                    j += N
            DIAGONALS[row][j] = MATRIX[i][j]
            DIAGONAL_TOTAL[row] += MATRIX[i][j]

    # 5. Memorise the fact that DIAGONALS arrays exist
    FLAG_MATRIX = True

#--------------------------------------------
def print_diagonals():
    """ """
    if not FLAG_MATRIX: 
        print("ERROR: you need to calculate DIAGONALS first")
        return
    print("Print Diagonals ...")
    print("       ",end="")
    for i in range(N):
        print("%c  "%(ALPHA[i]),end="")
    print()
    for i in range(N):
        print("%02d -- "%(i),end="")
        for j in range(N):
            print("%2d "%(DIAGONALS[i][j]),end="")
        print(": %2d"%(DIAGONAL_TOTAL[i]))
    print()
#--------------------------------------------
def print_matrix():
    """ """
    if not FLAG_MATRIX: 
        print("ERROR: you need to calculate MATRIX first")
        return
    print("Print Matrix ...")
    print("       ",end="")
    for i in range(N):
        print("%c  "%(ALPHA[i]),end="")
    print()
    for i in range(N):
        print("%02d -- "%(i),end="")
        for j in range(N):
            print("%2d "%(MATRIX[i][j]),end="")
        print()
        
#--------------------------------------------
def print_phi():
    """ """
    global FLAG_TOTAL
    if not FLAG_MATRIX: 
        print("ERROR: you need to calculate DIAGONALS first")
        return
    print("Print PHI of each diagonal...")
    print("Row #_of_letters Actual_phi random  no-random")
    for row in range(N):
        num_letters = num_of_letters(row)
        phi = calculate_phi(DIAGONALS[row])
        epr = 0.0385 * num_letters * (num_letters - 1)
        epp = 0.0683 * num_letters * (num_letters - 1)
        print("%2d  %4d           %4d     %4d     %4d" % \
            (row, num_letters, phi, int(epr+0.5), int(epp+0.5)))
        DIAGONAL_TOTAL[row] = num_letters
    FLAG_TOTAL = True
#--------------------------------------------
def match_two_diagonals():
    """ """
    if not FLAG_MATRIX: 
        print("ERROR: Create Diagonals first")
        return
    print("Matching two alphabets ...")
    a1 = int(input("Enter stationary alphabet (-1 for combined):"))
    a2 = int(input("Which alphabet to match:"))
    for i in range(N):
        chi = 0
        for j in range(N):
            a = i+j
            if a >= N: a -= N
            if a1 == -1:
                chi += COMBINED['stats'][j] \
                    *  DIAGONALS[a2][a]
            else:
                chi += DIAGONALS[a1][j] \
                    *  DIAGONALS[a2][a]
        print("%c:%4d"%(chr(i+ord('A')), chi),end=",")
    print()
    if a1 == -1:
        total_a1 = COMBINED['nblet']
    else:
        total_a1 = DIAGONAL_TOTAL[a1]
    print("Random chi    = ", int(0.0385 * total_a1 * \
        DIAGONAL_TOTAL[a2] ))
    print("Non-random chi= ", int(0.0684 * total_a1 * \
        DIAGONAL_TOTAL[a2] )) 
#--------------------------------------------
def remaining():
    """ Calculate PHI for remaining alphabets """
    if not FLAG_BASE: 
        print("ERROR: You must add at least one alphabet (Base alphabet)")
        return

    print("Alphabet Best_letter_match observed_chi random  no-random")
    for row in range(N):
        if row in COMBINED['histo']: continue
        letter = -1
        max_chi = 0
        for i in range(N):
            chi = 0
            for j in range(N):
                a = i+j 
                if a >= N: a -= N
                chi += COMBINED['stats'][j] \
                    *  DIAGONALS[row][a]
            if chi >= max_chi:
                max_chi = chi
                letter = i
        nblet1 = DIAGONAL_TOTAL[row]
        nblet2 = COMBINED['nblet'] 
        epr = int(0.0385 * nblet1 * nblet2)
        epp = int(0.0684 * nblet1 * nblet2) 
        print(("  %2d           %c              %4d" + \
            "       %4d        %4d") \
            % (row, chr(letter+ord('A')), max_chi, epr, epp))
        DECAL_HYPO[row] = letter
    print("alphabet: ",num2let(COMBINED['letter']))
    for i in range(N):
        if COMBINED['letter'][i] != -1:
            DECAL_HYPO[i] = COMBINED['letter'][i]
    print("Current hypothesis: ", num2let(DECAL_HYPO))
    print("Historic: ",COMBINED['histo'])
#--------------------------------------------
def combine():
    """ """
    global FLAG_BASE

    if not FLAG_MATRIX: 
        print("ERROR: you need to calculate DIAGONALS first")
        return
    print("Combine another alphabet to current combined alphabets...")
    if not FLAG_BASE:
        print("WARNING! You are ready to add the first alphabet (Base)")
        print("         its shift will be 'A'.")
    ch = input("Which alphabet to combine (enter to exit)?")
    if ch == "": return
    if not ch.isnumeric():
        return
    else:
        alphabet = int(ch)
        if alphabet < 0 or alphabet > 25:
            return
    if not FLAG_BASE:
        let_ch = "A"
        FLAG_BASE = True
    else:
        let_ch = input("Which letter [shift]?")
    let_int = ALPHA.find(let_ch)
    if COMBINED['letter'][alphabet] != -1:
        print("ERROR: This alphabet was already set")
        return
    COMBINED['letter'][alphabet] = let_int
    COMBINED['histo'].append(alphabet)
    COMBINED['nblet'] += DIAGONAL_TOTAL[alphabet]
    print("----> O: Old combined, A: Alphabet to combine, N: New combined")
    print("  ", end="")
    for i in range(N):
        print(" %c " % (ALPHA[i]), end="")
    print()
    print("O ", end="")
    for i in range(N):
        print("%2d " % (COMBINED['stats'][i]),end="")
    print()
    print("  ", end="")
    for i in range(N):
        j = let_int + i
        if j >= N: j -= N
        print(" %c " % (ALPHA[j]), end="")
    print()
    print("A ", end="")
    for i in range(N):
        j = let_int + i
        if j >= N: j -= N
        print("%2d " % (DIAGONALS[alphabet][j]), end="")
    print()
    print("N ", end="")
    for i in range(N):
        decalage = i + let_int
        if decalage >= N: decalage -= N
        COMBINED['stats'][i] += DIAGONALS[alphabet][decalage]
        print("%2d " % (COMBINED['stats'][i]),end="")
    print() 
#--------------------------------------------
def combine_reset():
    """ """
    global FLAG_BASE

    FLAG_BASE = False
    for i in range(N):
        COMBINED['stats'][i] = 0
    COMBINED['nblet'] = 0 
    COMBINED['histo'] = []
    COMBINED['letter'] = list(tb2)
    print("Combined data reseted...")
    
#--------------------------------------------
def combine_tout():
    """ """
    if not FLAG_MATRIX: 
        print("ERROR: you need to calculate DIAGONALS first")
        return
    print("Combine all alphabets after enter decalages...")
    print("WARNING: all values about combined alphabet are reset")
    pi = input("Decalage Alphabet? (return to exit, HYPO to use hypothesis)")
    if pi == "": return
    if pi == "HYPO":
        pi = num2let(DECAL_HYPO)
    if len(pi) != 26: return

    for i in range(N):
        COMBINED['stats'][i] = 0
    COMBINED['nblet'] = 0 

    for row in range(N):
        for i in range(N):
            let_int = ALPHA.find(pi[row])
            COMBINED['letter'][row] = let_int
            decalage = i + let_int
            if decalage >= N: decalage -= N
            COMBINED['stats'][i] += DIAGONALS[row][decalage]
        COMBINED['nblet'] += DIAGONAL_TOTAL[row]
    print("OK", COMBINED['nblet'])
#--------------------------------------------
def print_combine():
    """ """
    if not FLAG_MATRIX:
        print("ERROR: You must calculate Diagonal alphabets first")
        return
    ch = input("Enter the number of alphabet or return for combined alphabet ? ")
    if ch == "":
        if not FLAG_BASE: 
            print("ERROR: You must add at least one alphabet (Base alphabet)")
            return
    else:
        if not valide_num_26(ch):
            print("ERROR: incorrect number of alphabet")
            return
    if ch == "":    
        print("alphabet: ",num2let(COMBINED['letter']))
        print("Nb letters: ", COMBINED['nblet'])
        print("Hypothesis: ", num2let(DECAL_HYPO))
        print("Historic: ", COMBINED['histo'])
    else:
        print("Number of diagonal: ", ch)

    if ch == "":
        tb_stats = COMBINED['stats']
    else:
        row = int(ch)
        tb_stats = DIAGONALS[row]
    for i in range(N):
        print("  %c" % (ALPHA[i]), end="")
    print()
    total = 0
    for i in range(N):
        print("%3d" % (tb_stats[i]),end="")
        total += tb_stats[i]
    print() 
    print("Letters: ", total)
    epr = int(0.0385 * total * (total-1))
    epp = int(0.0684 * total * (total-1))
    max_chi = calculate_phi( tb_stats )
    print("Phi values: ")
    print("     Observed    = %10d" % (max_chi))
    print("     Random      = %10d" % (epr))
    print("     No-random   = %10d" % (epp))
#--------------------------------------------
def reduction():
    """ """
    if not FLAG_BASE: 
        print("ERROR: You must add at least one alphabet (Base alphabet)")
        return
    print("The Reduction Tableau ...")
    for i in range(N):
        if COMBINED['letter'][i] == -1: continue
        for j in range(N):
            row = i - j
            if row < 0: row += N
            col = j
            RED_TABLE[row][col] = (COMBINED['letter'][i] + j)%N
    print("   ", end="")
    for i in range(N):
        print(" %c " % (ALPHA[i]), end="")
    print()
    for i in range(N):
        print("%2d " % (i), end="")
        for j in range(N):
            print("%c  " % (ALPHA[RED_TABLE[i][j]]),end="")
        print()
    print()
#--------------------------------------------
def mono_alphabet():
    """ """
    for i in range(N):
        if COMBINED['letter'][i] == -1:
            print("ERROR: You need to discover all alphabets shifts")
            return
    diag_row = COMBINED['letter']
    nb_let = COMBINED['nblet']
    print("Nb of letters; ", nb_let, "Decalages: ", diag_row)
    
    pseudo_plain = ""
    for i in range(LG):
        if (i%N) == 0: print()
        crypt_letter =  CRYPTO[i]
        r = i % N  # position  du rotor
        if REVERSE:
            row = diag_row[ (r - crypt_letter) % N ] # la diagonale
        else:
            row = diag_row[ (r + crypt_letter) % N ] # la diagonale
        ddp = (row + r) % N  # positional value of the letter in the row 
        plain_letter = (i - ddp + crypt_letter) % N
        plain_letter_char = ALPHA[plain_letter]
        pseudo_plain += plain_letter_char
        print("%c " % (plain_letter_char), end="")
    print()
    print("Do you want to save mono-alphabet ciphertext ")
    ch = input("(return to exit, name of a file to save) ? ")
    if ch == "": return
    try:
        f = open(ch,"w")
        f.write(pseudo_plain)
        f.close()
    except:
        print("ERROR: (name of file incorrect, insufficient permissions, ...")
        return
    else:
        print("Plain saved ... ") 
#--------------------------------------------
#========> Principal loop <==================

menu()
while True:
    choice = input("Choice (return: print menu, Q: quit) ? ").upper()
    if   choice == "":
        menu()
    elif choice == 'Q':
        break
    elif choice == '1':
        lecture()
    elif choice == '2':
        print_cry()
    elif choice == '3':
        reverse_mode()
    elif choice == '4':
        create_diagonals()
    elif choice == '5':
        print_matrix()
    elif choice == '6':
        print_diagonals()
    elif choice == '7':
        print_phi()
    elif choice == '8':
        combine()
    elif choice == '9':
        remaining()
    elif choice == 'A':
        print_combine()
    elif choice == 'B':
        reduction()
    elif choice == 'C':
        combine_tout()
    elif choice == 'D':
        match_two_diagonals()
    elif choice == 'E':
        combine_reset()
    elif choice == 'F':
        mono_alphabet()

