# friedman_h1.py
# obj: 
#   Trouver les "basic cipher-text sequences"
#   a partir d'un cryptogramme. Ceci est un 
#   prealable a trouver la cle d'une machine Hebern
#   La methode des intervals utilisee a ete invente
#   par W.F. Friedman pour casser la Hebern 5.
#   Cette version est specifique de la machine Hebern 1
#   rotor. Elle correspond a la version v1 de la
#   methode des intervalles.

import sys
import string
import getopt

ALPHA = string.ascii_uppercase
N = 26
#======================================
def lit_stdin():
    entree = ""
    leTexte = sys.stdin.readlines()
    for i in range(len(leTexte)):
        entree += leTexte[i]
    flux_entre = ""
    for car in entree:
        if car.upper() in ALPHA:
            flux_entre += car.upper()
    return flux_entre
#======================================
def usage():
    print("usage:")
    print("Syntax : friedman_h1.py [options...]")
    print("Example: echo AAAAA | python friedman.py -I 5 ")
    print("Options:")
    print("  -I interval Interval (5 by default),")
    print("              if I==5, then X=0,....Y=5, then")
    print("              X and Y are separated by 4 letters")
    print("  -K key      The External Key (A by default)")

#======================================
try:
  if len(sys.argv) < 1:
    raise getopt.GetoptError("too few arguments")
  opts, args = getopt.getopt(sys.argv[1:],
    "hI:K:", ["help","interval=", "key="])
except getopt.GetoptError as err:
  # print help information and exit:
  print(sys.argv[0],": ", str(err))# option -a not recognized"
  usage()
  sys.exit(2)

DEBUG = False
KEY = 'A'
INTERVAL = 5

for o, a in opts:
  if   o in ("-D", "--debug"):
    DEBUG = True
  elif o in ("-h", "--help"):
    usage()
    sys.exit()
  elif o in ("-K", "--key"):
    if a in ALPHA:
        KEY = a
  elif o in ("-I", "--interval"):
    if a.isnumeric():
        INTERVAL = int(a)
  else:
    assert False, "unhandled option"

#--------------------------------------
CRYPTO = lit_stdin()
DEBUT = ALPHA.find(KEY)
i = DEBUT
STATS = []
tb = [0]*N
for k in range(N):
    STATS.append(list(tb))
while True:
    if (i+INTERVAL) >= len(CRYPTO):
        break
    let1 = ALPHA.find(CRYPTO[i])
    let2 = ALPHA.find(CRYPTO[i+INTERVAL])
    STATS[let1][let2] += 1
    i += N

print("   ", end="")
for i in range(N):
    print("%2c" % (ALPHA[i]), end=" ")
print()
for i in range(N):
    print("%c:" % (ALPHA[i]), end=" ")
    for j in range(N):
        print("%2d" % (STATS[j][i]), end=" ")
    print()

TOTAL = list(tb)
for i in range(N):
    for j in range(N):
        TOTAL[i] += STATS[i][j]
print("  ", end="")
for i in range(N):
    print("%3d" % (TOTAL[i]), end="")
print()

