# hebern1_table.py 
# obj: a software for create tables to cipher/decipher
#      theses tables emulate a Hebern cipher machine
#------------------------------------------------------
# Todo list
# - add the ciphering with Lampboard and Keyboard
# - Possibility to use Inversed rotor

import sys
import string
import getopt

#======================================
def lit_rotor(nom):
    f = open("ROTORS/" + nom + ".rot")
    permutation = f.readline()
    f.close()
    return permutation.strip()
#======================================
#======================================
def table_cipher():
    print("Rotor in clockwise direction, Coord of a ciphered letter: ")
    print("Abscissa: Plain letter, ordinate: Position of rotor (key)")
    print()
    print("  A B C D E F G H I J K L M N O P Q R S T U V W X Y Z")
    for i in range(N):
        print(ALPHA[i], end=" ")
        for j in range(N):
            y = KEY_PI[j]
            y = (PI[ (y+i)%N ] - i )%N
            y = LAMP_PI[y]
            print(ALPHA[y], end=" ")
        print()
    print()
#======================================
def table_decipher():
    print("Rotor in clockwise direction, Coord of a plain letter: ")
    print("Abscissa: Ciphered letter, ordinate: Position of rotor (key)")
    print()
    print("  A B C D E F G H I J K L M N O P Q R S T U V W X Y Z")
    for i in range(N):
        print(ALPHA[i], end=" ")
        for j in range(N):
            y = LAMP_inv[j]
            y = (PI_inv[ (y+i)%N ] - i )%N
            y = KEY_inv[y]
            print(ALPHA[y], end=" ")
        print()
    print()
#======================================
def table_trigo_cipher():
    print("Rotor in trigonometric direction, Coord of a ciphered letter: ")
    print("Abscissa: Plain letter, ordinate: Position of rotor (key)")
    print()
    print("  A B C D E F G H I J K L M N O P Q R S T U V W X Y Z")
    for i in range(N):
        print(ALPHA[i], end=" ")
        for j in range(N):
            y = KEY_PI[j]
            y = (PI[ (y-i)%N ] + i )%N
            y = LAMP_PI[y]
            print(ALPHA[y], end=" ")
        print()
    print()
#======================================
def table_trigo_decipher():
    print("Rotor in trigonometric direction, Coord of a plain letter: ")
    print("Abscissa: Deciphered letter, ordinate: Position of rotor (key)")
    print()
    print("  A B C D E F G H I J K L M N O P Q R S T U V W X Y Z")
    for i in range(N):
        print(ALPHA[i], end=" ")
        for j in range(N):
            y = LAMP_inv[j]
            y = (PI_inv[ (y-i)%N ] + i )%N
            y = KEY_inv[y]
            print(ALPHA[y], end=" ")
        print()
    print()
#======================================
def rot2int(un_rotor):
    """ revoie un rotor sous forme d'une suite d'entiers """
    tb = []
    for c in un_rotor:
        x = ALPHA.find(c)
        tb.append(x)
    return tb
#======================================
def invrot(un_rotor):
    """ inverse un rotor """
    PIinv = [0]*N
    for x in range(N):
        y = un_rotor[x]
        PIinv[y] = (un_rotor[x] - (y - x))%N
    return PIinv
#======================================
def usage():
    print("usage:")
    print("Syntax : hebern1_table.py [options...]")
    print("Options:")
    print("  -L lamp     The lampboard permutation (iden by defaults)")
    print("  -K keyb     The keyboard permutation (iden by defaults)")
    print("  -R rotor    The rotor (iden  by defaults)")
    print("  -I          Use the Rotor in inverse mode")
    print("  -d          Decipher Mode (cipher mode by default)")
    print("  -t          Trigonometric direction (clockwise by default)")
    print("Note: for L,K,R options, if you begins with =, the wiring is")
    print("given, otherwise, you must give a file which contains the wiring")
#======================================
try:
  opts, args = getopt.getopt(sys.argv[1:],
    "hL:K:R:Idt", ["help","lamp=","keyb=","rotor=",
    "inverse", "decipher", "trigo" ])
except getopt.GetoptError as err:
  # print help information and exit:
  print(sys.argv[0],": ", str(err))# option -a not recognized"
  usage()
  sys.exit(2)

N = 26
ALPHA = string.ascii_uppercase
DECIPHER_MODE = False
INVERSE = False
TRIGO_DIR = False
STATORKEY = ALPHA
ROTOR     = ALPHA
STATORLAMP= ALPHA

for o, a in opts:
  if   o in ("-h", "--help"):
    usage()
    sys.exit()
  elif o in ("-L", "--lamp"):
    if a[0] == "=":
        STATORLAMP = a[1:]
    else:
        STATORLAMP = lit_rotor(a)
  elif o in ("-K", "--keyb"):
    if a[0] == "=":
        STATORKEY = a[1:]
    else:
        STATORKEY = lit_rotor(a)
  elif o in ("-R", "--rotor"):
    if a[0] == "=":
        ROTOR = a[1:]
    else:
        ROTOR = lit_rotor(a)
  elif o in ("-E", "--external="):
    KEYEXTERN = a
  elif o in ("-d", "--decipher"):
    DECIPHER_MODE = True 
  elif o in ("-t", "--trigo"):
    TRIGO_DIR = True
  else:
    assert False, "unhandled option"

#--------------------------------------
PI = rot2int( ROTOR )
PI_inv = invrot( PI )
print("PI: ", PI)
print("PI_inv: ", PI_inv)
KEY_PI = rot2int( STATORKEY )
LAMP_PI = rot2int( STATORLAMP )
KEY_inv = invrot( KEY_PI )
LAMP_inv = invrot( LAMP_PI )

if DECIPHER_MODE:
    if TRIGO_DIR:
        table_trigo_decipher()
    else:
        table_decipher()
else:
    if TRIGO_DIR:
        table_trigo_cipher()
    else:
        table_cipher()

