# konheim.py
# utilise l'algorithme de Konheim
# "Computer Security and Cryptography", by Alan G. Konheim
# Chapter 6.5 Analysis of A 1-Rotor system: ciphertext only (p156)
# ------------------------------------------------------
# Chi[t,s] = probability of the plain letter 't' is ciphered by 's'
#            when the key (i) is zero.
# NL[s][i] = Number of letter 's'-i at the "i" key
# N[i]     = Number of letter ciphered at the "i" key
# PI[x]    = Probability of the plain letter "x"
# ------------------> formula: 
# Chi[t,s] = SUM_FROM_i=0_TO_i=25 (PI[t-i] * NL[s][i] / N[i])

from ROTOR import *
import sys
import string
ALPHA = string.ascii_uppercase

# ==================================
def litfic( fic ):
    f = open(fic,"r")
    leTexte = f.readlines()
    texteConcat = ""
    f.close()
    for i in range(len(leTexte)):
        ch = leTexte[i]
        ch = ch.upper()
        for j in range(len(ch)):
            if not(ch[j] in ALPHA):
                continue
            texteConcat = texteConcat + ch[j]
    return texteConcat
# ==================================
def litproba( fic ):
    tb = [0]*NA
    f = open("STATS/" + fic + ".proba","r")
    while True:
        ligne = f.readline()
        if ligne == "": break
        lettre, proba = ligne.split(",")
        num_letter = let2idx(lettre)
        tb[num_letter] = float(proba)
    return tb
# ==================================
def let2idx( letter ):
    return ALPHA.index( letter.upper() )
# ==================================
if len(sys.argv) < 2:
    print("konheim: search of probability of a plain letter ")
    print("         is ciphered by a specific letter at key 0")
    print("Syntax: %s language cryptogram plain_letter" \
        % (sys.argv[0]))
    sys.exit(1) 

NA = 26
PI = litproba( sys.argv[1] )
crypto = litfic(sys.argv[2])
lettre_clair = sys.argv[3]
PLAIN_LETTER = let2idx( lettre_clair )

print("Plain letters probability: ", PI[0:5], "...")
print("cryptogram : ", crypto[0:50], "...")
print("Cryptogram Length : ", len(crypto))
print("Plain letter: ", ALPHA[ PLAIN_LETTER ])

if len(crypto) < 400:
    print("ERROR: the cryptogram is too short")
    sys.exit(1)
N = [0] * NA
NL = []
for i in range(NA):
    NL.append(list([0]*NA))

for j in range(len(crypto)):
    #print("<", crypto[j], ">", end="")
    i = j % NA
    N[i] += 1
    idx = let2idx( crypto[j] )
    NL[ idx ][ i ]  += 1   
#print()
#print("N[]:", N)
#for i in range(NA):
#    print(NL[i])
for let_cry in range(NA):
    chi = 0
    for i in range(NA):
        chi += PI[(PLAIN_LETTER-i)%NA] \
           * (NL[(let_cry-i)%NA][i] / N[i])
    print("%c -> %c %6.4f\t " % \
        (ALPHA[PLAIN_LETTER],ALPHA[let_cry].lower(), chi), end="")
    if ((let_cry+1) % 5) == 0:
        print()
print()
