# motprobable.py 
# obj: rechercher si un mot probable est present 
#      et si oui ou. On utilise la methode des isomorphe
#      il faut connaitre le keyboard et le rotor
#      le lampboard est inconnu.
#      on peut utiliser ce programme pour trouver
#      le rotor utilise (si on connait plusieurs rotors)

import sys
from Hebern1Rot import *
import string
import getopt

#======================================
def is_isomorph(ch1,ch2):
    dico1 = {}
    dico2 = {}
    for i in range(len(ch1)):
        c1 = ch1[i]
        c2 = ch2[i]

        if c1 in dico1:
            val = dico1[c1]
            if val != c2:
                return False
        else:
            dico1[c1] = c2

        if c2 in dico2:
            val = dico2[c2]
            if val != c1:
                return False
        else:
            dico2[c2] = c1
    return True
#======================================
def lit_rotor(nom):
    f = open("ROTORS/" + nom + ".rot")
    permutation = f.readline()
    f.close()
    return permutation.strip()    
#======================================
def lit_stdin():
    entree = ""
    leTexte = sys.stdin.readlines()
    for i in range(len(leTexte)):
        entree += leTexte[i]
    flux_entre = ""
    for car in entree:
        if car.upper() in ALPHA:
            flux_entre += car.upper()
    return flux_entre
#======================================
def usage():
    print("usage:")
    print("Syntax : motprobable.py [options...]")
    print("Example: echo AAAAA | python motprobable.py -R CW1 ")
    print("         the Internal Key file need to end by .key")
    print("Options:")
    print("  -C crib     The probable word (crib)")
    print("  -K keyb     The keyboard permutation (iden by defaults)")
    print("  -R rotor    The rotor (iden  by defaults)")
    print("  -t          Trigonometric direction (False [clockwise] by default")
    print("  -s          Scan the cryptogram (search the crib every where")
    print("              by default, only inspect the begin of the crypto")
    print("  -D          Debug Mode")
    print("Note: for K,R options, if you begins with =, the wiring is")
    print("given, otherwise, you must give a file which contains the wiring")

#======================================
try:
  if len(sys.argv) == 1:
    raise getopt.GetoptError("too few arguments")
  opts, args = getopt.getopt(sys.argv[1:],
    "hK:R:C:Dts", ["help","keyb=","rotor=", 
    "crib=","debug","trigo","scan" ])
except getopt.GetoptError as err:
  # print help information and exit:
  print(sys.argv[0],": ", str(err))# option -a not recognized"
  usage()
  sys.exit(2)

DEBUG = False
TRIGO_MODE = False
SCAN = False
STATORKEY = string.ascii_uppercase 
ROTOR     = string.ascii_uppercase 
STATORLAMP= string.ascii_uppercase 
CIPHER_MODE= True
REVERSE = False
CRIB = ""
LEN_CRIB_MINI = 12

for o, a in opts:
  if   o in ("-D", "--debug"):
    DEBUG = True
  elif o in ("-h", "--help"):
    usage()
    sys.exit()
  elif o in ("-K", "--keyb"):
    if a[0] == "=":
        STATORKEY = a[1:]
    else:
        STATORKEY = lit_rotor(a)
  elif o in ("-R", "--rotor"):
    if a[0] == "=":
        ROTOR = a[1:]
    else:
        ROTOR = lit_rotor(a)
  elif o in ("-t", "--trigo"):
    TRIGO_MODE = True
  elif o in ("-s", "--scan"):
    SCAN = True
  elif o in ("-C", "--crib"):
    CRIB = a
  else:
    assert False, "unhandled option"

if len(CRIB) < LEN_CRIB_MINI:
    print("Sorry, the Crib is too short. It is less than ", \
        LEN_CRIB_MINI, "charaters.")
    sys.exit(1)
#--------------------------------------
c = HEBERN( ROTOR, STATORKEY,STATORLAMP,CIPHER_MODE,TRIGO_MODE,REVERSE,DEBUG)
if DEBUG : 
    print(c.info())

crypto = lit_stdin()
isomorph = "" 
n = 0
idx = 1
# ---- calculate the cipherment of the probable word
#      for all 26 possible keys without lampboard
pseudo_plain = []
for i in range(N):
    c.Rotor.rewind()
    for j in range(i):
        c.move()
    ch = ""
    for car in CRIB:
        car2 = c.cipher_a_char(car)
        ch += car2
    pseudo_plain.append( ch )

# ---- scan ciphertext
if SCAN:
    MAX = len(crypto)
else:
    MAX = len(CRIB)
print("Length of the cryptogram: ", len(crypto))
print("The last position tested: ", MAX-len(CRIB))
print("(if the last position tested is zero, we test only")
print(" the beginning of the cryptogram [we don't use -s])")
print("1st col: The current position (inside the cryptogram)")
print("2nd col: The Key (for the 1st character of the isomorph)")
print("3th col: The slice of the cryptogram tested")
print("4th col: The pseudo plain which corresponds to the")
print("slice of the cryptogram tested, in short, the isomorph")
for i in range(MAX-len(CRIB)+1):
    isomorph = crypto[i:i+len(CRIB)]
    for j in range(N):
        code = is_isomorph(pseudo_plain[j], isomorph)
        if code:
            print("%04d " % (i), end=", ")
            print(ALPHA[j], isomorph, pseudo_plain[j])

