# signature.py 
# obj: a software to identify a rotor
#------------------------------------------------------
# Todo list
# - manage permutation type 

import sys
import string
import getopt
from ROTOR import *

#======================================
def lit_rotor(nom):
    f = open("ROTORS/" + nom + ".rot")
    permutation = f.readline()
    f.close()
    return permutation.strip()
#======================================
def recherche(tb):
    digits = 1
    while True:
        petits = list([0] * N)
        for i in range(N):
            idx = 0
            for j in range(digits):
                v = (tb[(i+j)%N] * (N**(digits-j-1)))
                idx += v
            petits[i] = idx
        plus_petit = min(petits)
        nb_petits = 0
        idx = -1
        for i in range(N):
            x = petits[i]
            if x == plus_petit:
                nb_petits += 1
                idx = i
        if nb_petits == 1:
            return idx
        else:
            digits += 1
        if digits > 5:
            print("... There is a problem")
            sys.exit(1)

#======================================
def diff_shift(un_shift):
    les_diff = []
    avant = un_shift[0]
    tb = un_shift[1:]
    tb.append(avant)
    for x in tb:
        une_diff = (x - avant) % N
        les_diff.append( une_diff )
        avant = x
    x = recherche(les_diff)
    les_diff = les_diff[x:]+les_diff[:x]
    return les_diff
#======================================
def permu_type(ch_rot):
    flg = [False]*N
    types = []
    in_cycle = False
    cpt = 0
    idx = 0
    while cpt < N:
        if in_cycle:
            if ALPHA[idx] == debut:
                in_cycle = False
                types.append(list(cycle))
                cpt += 1
            else:
                cycle.append(ALPHA[idx])
                flg[idx] =  True
                idx = ALPHA.find(ch_rot[idx])
        else:
            if flg[cpt]:
                cpt += 1
                continue
            debut = ALPHA[cpt]
            cycle = [debut]
            in_cycle = True
            idx = ALPHA.find(ch_rot[cpt])
    types = sorted(types, key=lambda x:len(x))
    chtyp = ""
    for un_cycle in types:
        chtyp += "(" + "".join(un_cycle) + "),"
    #print(">>>", chtyp)
    tb = []
    for un_cycle in types:
        tb.append(len(un_cycle))
    tb.sort()
    ch = ""
    for x in tb:
        ch += str(x) + "."
    types.sort()
    return [ch, chtyp]

#======================================
#======================================
def rot2int(un_rotor):
    """ revoie un rotor sous forme d'une suite d'entiers """
    tb = []
    for c in un_rotor:
        x = ALPHA.find(c)
        tb.append(x)
    return tb
#======================================
def invrot(un_rotor):
    """ inverse un rotor """
    PIinv = [0]*N
    for x in range(N):
        y = un_rotor[x]
        PIinv[y] = (un_rotor[x] - (y - x))%N
    return PIinv
#======================================
def usage():
    print("usage:")
    print("Syntax : signature.py [options...]")
    print("Options:")
    print("  -R rotor    The rotor (identity by defaults)")
    print("  -r          Print information about reverse rotor")
    print("  -c          Print information about cycles")
    print("  -S          Print information about signature")
    print("  -s          Print the complete signature")
    print("By default, print only the beginning of the signature")
    print("  -p          print the wiring (PI)")
    print("  -a          Print all (by default)")
    print("Note: for the R option, if you begins with =, the wiring is")
    print("given, otherwise, you must give a file which contains the wiring")
    print("This file must be in ROTORS directory and must have *.rot extension")
#======================================
try:
  opts, args = getopt.getopt(sys.argv[1:],
    "hL:R:rcSspa", ["help","lamp=","rotor=","reverse","cycles",
    "sign","signature","pi","all"])
except getopt.GetoptError as err:
  # print help information and exit:
  print(sys.argv[0],": ", str(err))# option -a not recognized"
  usage()
  sys.exit(2)

N = 26
ALPHA = string.ascii_uppercase
INVERSE = False
MYROTOR = ALPHA
SIGN_COMP = False
REVERSE = False
CYCLES = False
SIGNATURE = False
PRINT_PI = False
ALL = False

for o, a in opts:
  if   o in ("-h", "--help"):
    usage()
    sys.exit()
  elif o in ("-R", "--rotor"):
    if a[0] == "=":
        MYROTOR = a[1:]
    else:
        MYROTOR = lit_rotor(a)
  elif o in ("-r", "--reverse"):
    REVERSE = True
  elif o in ("-s", "--sign"):
    SIGN_COMP = True
  elif o in ("-c", "--cycles"):
    CYCLES = True
  elif o in ("-S", "--signature"):
    SIGNATURE = True
  elif o in ("-p", "--pi"):
    PRINT_PI = True
  elif o in ("-a", "--all"):
    REVERSE = True
    CYCLES  = True
    PRINT_PI= True
    SIGNATURE = True
  else:
    assert False, "unhandled option"

#--------------------------------------
if not SIGNATURE and not CYCLES:
    SIGNATURE = True
if MYROTOR == ALPHA:
    print("WARNING: use -h option to print syntax")
PI = MYROTOR
if not Rotor.valide(PI):
    print("Rotor wiring is not correct")
    sys.exit(1)
c = Rotor(PI)
les_rot = [
        [MYROTOR,"PI direct"],
        [Rotor.rotor2char(Rotor.decal2pi(c.PI_SHIFT_INV)),"PI reverse"],
]        

for couple in les_rot:
    PI,info = couple
    c = Rotor(PI)
    if REVERSE: print(info)
    if PRINT_PI:
        print("  PI       : ", PI)
    if CYCLES:
        le_type, les_cycles = permu_type(PI)
        print("  Type     : ", le_type)
        print("  Cycles   : ", les_cycles)
    if SIGNATURE:
        if SIGN_COMP:
            print("  Signature: ", diff_shift(c.PI_SHIFT))
        else:
            print("  Signature: ", diff_shift(c.PI_SHIFT)[:10])
    if not REVERSE: break
    
